from typing import Optional, List

import numpy as np

from hydtank.basin_def import BasinDef
from hydtank.parameters import SubbasinParameters


class Subbasin(BasinDef):
    def __init__(
            self, name: str,
            area: float,
            downstream: Optional[BasinDef] = None,
            upstream: Optional[List[BasinDef]] = None,
            parameters: SubbasinParameters = SubbasinParameters()
    ):
        super().__init__(name, downstream, upstream)

        self._area = area
        self._parameters = parameters

        self._q_tank_0: Optional[np.ndarray] = None
        self._q_tank_1: Optional[np.ndarray] = None
        self._q_tank_2: Optional[np.ndarray] = None
        self._q_tank_3: Optional[np.ndarray] = None

        self._side_outlet_flow_tank_0: Optional[np.ndarray] = None
        self._side_outlet_flow_tank_1: Optional[np.ndarray] = None
        self._side_outlet_flow_tank_2: Optional[np.ndarray] = None
        self._side_outlet_flow_tank_3: Optional[np.ndarray] = None

        self._bottom_outlet_flow_tank_0: Optional[np.ndarray] = None
        self._bottom_outlet_flow_tank_1: Optional[np.ndarray] = None
        self._bottom_outlet_flow_tank_2: Optional[np.ndarray] = None

    @property
    def bottom_outlet_flow_tank_0(self) -> Optional[np.ndarray]:
        return self._bottom_outlet_flow_tank_0

    @property
    def bottom_outlet_flow_tank_1(self) -> Optional[np.ndarray]:
        return self._bottom_outlet_flow_tank_1

    @property
    def bottom_outlet_flow_tank_2(self) -> Optional[np.ndarray]:
        return self._bottom_outlet_flow_tank_2

    @property
    def side_outlet_flow_tank_1(self) -> Optional[np.ndarray]:
        return self._side_outlet_flow_tank_1

    @property
    def side_outlet_flow_tank_2(self) -> Optional[np.ndarray]:
        return self._side_outlet_flow_tank_2

    @property
    def side_outlet_flow_tank_3(self) -> Optional[np.ndarray]:
        return self._side_outlet_flow_tank_3

    @property
    def q_tank_0(self) -> Optional[np.ndarray]:
        return self._q_tank_0

    @property
    def q_tank_1(self) -> Optional[np.ndarray]:
        return self._q_tank_1

    @property
    def q_tank_2(self) -> Optional[np.ndarray]:
        return self._q_tank_2

    @property
    def q_tank_3(self) -> Optional[np.ndarray]:
        return self._q_tank_3

    @property
    def area(self):
        return self._area

    def tank_discharge(self, p: np.ndarray, e: np.ndarray, time_step: int, interval: float) -> List[str]:
        logs = []
        params = self._parameters

        if params.t0_soh_uo < params.t0_soh_lo:
            warnings_msg = f'WARNING-TANK-01 ({self.name}): Invalid parameter upper outlet height is less than lower outlet height (Tank 0)'
            logs.append(warnings_msg)
            print(warnings_msg)

        tank_storage = np.zeros((time_step, 4), dtype=np.float64)
        side_outlet_flow = np.zeros((time_step, 4), dtype=np.float64)
        bottom_outlet_flow = np.zeros((time_step, 3), dtype=np.float64)

        del_rf_et = p - e

        tank_storage[0, 0] = max(params.t0_is, 0)
        tank_storage[0, 1] = max(params.t1_is, 0)
        tank_storage[0, 2] = max(params.t2_is, 0)
        tank_storage[0, 3] = max(params.t3_is, 0)

        for t in np.arange(time_step):
            # TANK 0 : surface runoff
            side_outlet_flow[t, 0] = params.t0_soc_lo * max(tank_storage[t, 0] - params.t0_soh_lo, 0) \
                                     + params.t0_soc_uo * max(tank_storage[t, 0] - params.t0_soh_uo, 0)

            # TANK 1 : intermediate runoff
            side_outlet_flow[t, 1] = params.t1_soc * max(tank_storage[t, 1] - params.t1_soh, 0)
            # TANK 2 : sub-base runoff
            side_outlet_flow[t, 2] = params.t2_soc * max(tank_storage[t, 2] - params.t2_soh, 0)
            # TANK 3 : base-flow | Side outlet height = 0
            side_outlet_flow[t, 3] = params.t3_soc * tank_storage[t, 3]

            bottom_outlet_flow[t, 0] = params.t0_boc * tank_storage[t, 0]
            bottom_outlet_flow[t, 1] = params.t1_boc * tank_storage[t, 1]
            bottom_outlet_flow[t, 2] = params.t2_boc * tank_storage[t, 2]

            if t < (time_step - 1):
                tank_storage[t + 1, 0] = tank_storage[t, 0] + del_rf_et[t + 1] - (
                        side_outlet_flow[t, 0] + bottom_outlet_flow[t, 0])

                tank_storage[t + 1, 1] = tank_storage[t, 1] + bottom_outlet_flow[t, 0] - (
                        side_outlet_flow[t, 1] + bottom_outlet_flow[t, 1])

                tank_storage[t + 1, 2] = tank_storage[t, 2] + bottom_outlet_flow[t, 1] - (
                        side_outlet_flow[t, 2] + bottom_outlet_flow[t, 2])

                tank_storage[t + 1, 3] = tank_storage[t, 3] + bottom_outlet_flow[t, 2] - side_outlet_flow[t, 3]

                tank_storage[t + 1, 0] = max(tank_storage[t + 1, 0], 0)
                tank_storage[t + 1, 1] = max(tank_storage[t + 1, 1], 0)
                tank_storage[t + 1, 2] = max(tank_storage[t + 1, 2], 0)
                tank_storage[t + 1, 3] = max(tank_storage[t + 1, 3], 0)

            for i in range(4):
                total_tank_outflow = bottom_outlet_flow[t, i] + side_outlet_flow[t, i] if i <= 2 else side_outlet_flow[
                    t, i]

                if total_tank_outflow > tank_storage[t, i]:
                    warnings_msg = f'WARNING-TANK-02 ({self.name}): Total outlet flow exceeded tank storage for tank {i} at timestep {t}'
                    logs.append(warnings_msg)
                    print(warnings_msg)

        unit_conv_coeff = (self.area * 1000) / (interval * 3600)
        discharge = unit_conv_coeff * side_outlet_flow.sum(axis=1)

        # Set result
        self._q_tank_0 = tank_storage[:, 0]
        self._q_tank_1 = tank_storage[:, 1]
        self._q_tank_2 = tank_storage[:, 2]
        self._q_tank_3 = tank_storage[:, 3]

        self._side_outlet_flow_tank_0 = side_outlet_flow[:, 0]
        self._side_outlet_flow_tank_1 = side_outlet_flow[:, 1]
        self._side_outlet_flow_tank_2 = side_outlet_flow[:, 2]
        self._side_outlet_flow_tank_3 = side_outlet_flow[:, 3]

        self._bottom_outlet_flow_tank_0 = bottom_outlet_flow[:, 0]
        self._bottom_outlet_flow_tank_1 = bottom_outlet_flow[:, 1]
        self._bottom_outlet_flow_tank_2 = bottom_outlet_flow[:, 2]

        self._q_sim = discharge

        return logs



from datetime import datetime
from io import StringIO
from typing import Optional

from hydutils.hyd_constants import (
    TIMESERIES,
    PRECIPITATION,
    EVAPOTRANSPIRATION,
    DISCHARGE,
)
import pandas as pd

from hydtank.basin_def_utils import (
    extract_basin_file_data,
    extract_basin_defs_dict,
    build_basin_defs,
    build_root_node,
)
from hydtank.dataset import Dataset
from hydtank.hydtank import HydTANK


def _build_basin_defs_from_content(basin_content: str):
    data = extract_basin_file_data(basin_content)

    basin_defs_dict = extract_basin_defs_dict(data)

    basin_defs = build_basin_defs(basin_defs_dict)

    return basin_defs


def build_hydtank_from_dataset(
    basin_content: str,
    _dataset: Dataset,
    interval: float = 24.0,
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
):
    basin_defs = _build_basin_defs_from_content(basin_content)

    root_node = build_root_node(basin_defs)

    return HydTANK(_dataset, basin_defs, root_node, interval, start, end)


def build_hydtank(
    dataset_content: str,
    basin_content: str,
    interval: float = 24.0,
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    timeseries_column_name: str = TIMESERIES,
    precipitations_column_name: str = PRECIPITATION,
    evapotranspirations_column_name: str = EVAPOTRANSPIRATION,
    discharge_column_name: str = DISCHARGE,
):
    df = pd.read_csv(StringIO(dataset_content))

    _dataset = Dataset(
        time_series=df[timeseries_column_name].tolist(),
        precipitation=df[precipitations_column_name].tolist(),
        evapotranspiration=df[evapotranspirations_column_name].tolist(),
        discharge=df[discharge_column_name].tolist(),
    )

    return build_hydtank_from_dataset(basin_content, _dataset, interval, start, end)



from typing import Optional, List

import numpy as np

from hydtank.basin_def import BasinDef
from hydtank.parameters import ReachParameters


class Reach(BasinDef):
    def __init__(
            self, name: str,
            downstream: Optional[BasinDef] = None,
            upstream: Optional[List[BasinDef]] = None,
            parameters: ReachParameters = ReachParameters()
    ):
        super().__init__(name, downstream, upstream)
        self._parameters = parameters

    def muskingum(
            self, n_step: int, interval: float
    ):
        logs = []
        sum_node = np.zeros(n_step, dtype=np.float64)
        params = self.parameters

        for us_node in self.upstream:
            inflow = us_node.q_sim
            outflow: np.ndarray = np.zeros(n_step, dtype=np.float64)

            c0: float = (-params.k * params.x + 0.5 * interval) / (
                    params.k * (1 - params.x) + 0.5 * interval)
            c1: float = (params.k * params.x + 0.5 * interval) / (
                    params.k * (1 - params.x) + 0.5 * interval)
            c2: float = (params.k * (1 - params.x) - 0.5 * interval) / (
                    params.k * (1 - params.x) + 0.5 * interval)

            if (c0 + c1 + c2) > 1 or params.x > 0.5 or (interval / params.k + params.x) > 1:
                warning_msg = f"WARNING-MUSKINGUM-01 ({self.name}): violates k, x constraints ({params.k}, {params.x})"
                logs.append(warning_msg)
                print(warning_msg)

            outflow[0] = inflow[0]

            for t in np.arange(1, n_step):
                outflow[t] = c0 * inflow[t] + c1 * inflow[t - 1] + c2 * outflow[t - 1]

            sum_node += outflow

        self._q_sim = sum_node

        return logs



import dataclasses
import json

import numpy as np


def mse(x: np.ndarray, y: np.ndarray):
    return ((x - y) ** 2).sum() / x.shape[0]


def rmse(x: np.ndarray, y: np.ndarray):
    return np.sqrt(mse(x, y))


def nse(sim: np.ndarray, obs: np.ndarray):
    obs_mean = obs.mean()
    return 1 - (np.square(obs - sim).sum() / np.square(obs - obs_mean).sum())


def r2(x: np.ndarray, y: np.ndarray):
    n = x.shape[0]

    nu = (n * ((x * y).sum()) - (x.sum()) * (y.sum())) ** 2
    de = (n * ((x ** 2).sum()) - (x.sum()) ** 2) * (n * ((y ** 2).sum()) - (y.sum()) ** 2)

    return nu / de


def pbias(obs: np.ndarray, sim: np.ndarray):
    return (obs - sim).sum() * 100 / obs.sum()


@dataclasses.dataclass
class Statistics:
    rmse: float = None
    nse: float = None
    r2: float = None
    pbias: float = None

    def calculate_stats(self, q_obs: np.ndarray, q_sim: np.ndarray):
        self.rmse = rmse(q_obs, q_sim)
        self.nse = nse(q_sim, q_obs)
        self.r2 = r2(q_sim, q_obs)
        self.pbias = pbias(q_sim, q_obs)

    def __str__(self):
        return json.dumps(self.__dict__, indent=4)



import copy
import io
import uuid
from datetime import datetime
from queue import Queue
from typing import Optional, List, Literal, Union

from hydutils.df_validation import (
    validate_columns_for_nulls,
    validate_interval,
    filter_timeseries,
)
from hydutils.hyd_constants import (
    TIMESERIES,
    INTERVAL,
    PRECIPITATION,
    EVAPOTRANSPIRATION,
    DISCHARGE,
)
import networkx as nx
import numpy as np
import pandas as pd
from PIL import Image
from matplotlib import pyplot as plt
from scipy.optimize import minimize

from hydtank.basin_def import BasinDef
from hydtank.bsd_junction import Junction
from hydtank.bsd_reach import Reach
from hydtank.bsd_sink import Sink
from hydtank.bsd_subbasin import Subbasin
from hydtank.dataset import Dataset
from hydtank.parameters import Parameters, SubbasinParameters, ReachParameters


class HydTANK:
    def __init__(
            self,
            dataset: Dataset,
            basin_defs: List[BasinDef],
            root_node: List[BasinDef],
            interval: float = 24.0,
            start: Optional[datetime] = None,
            end: Optional[datetime] = None,
    ):
        self._dataset = dataset
        self._basin_defs = basin_defs
        self._root_node = root_node

        self._interval = interval
        self._start = start
        self._end = end

        self._timeseries: Optional[np.ndarray] = None
        self._P: Optional[np.ndarray] = None
        self._E: Optional[np.ndarray] = None
        self._q_obs: Optional[np.ndarray] = None

        self._name = uuid.uuid4().hex

        self._logs: List[str] = []

        self._run()

    @property
    def dataset(self):
        return self._dataset

    @property
    def basin_defs(self):
        return self._basin_defs

    @property
    def root_node(self):
        return self._root_node

    @property
    def interval(self):
        return self._interval

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def logs(self):
        return self._logs

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    def copy(self):
        return copy.deepcopy(self)

    def _validate_dataset_n_provide_dataframe(self):
        df = self._dataset.to_dataframe()
        df = validate_columns_for_nulls(df)
        df = validate_interval(df, self._interval)
        return df

    def _init_input(self, df: pd.DataFrame):
        self._timeseries = df[TIMESERIES].reset_index(drop=True).to_numpy()
        self._P = df[PRECIPITATION].reset_index(drop=True).to_numpy()
        self._E = df[EVAPOTRANSPIRATION].reset_index(drop=True).to_numpy()
        self._q_obs = df[DISCHARGE].reset_index(drop=True).to_numpy()

    def _build_computation_stack(self):
        computation_stack = []
        node_queue: Queue[BasinDef] = Queue()
        for root_node in self._root_node:
            node_queue.put(root_node)

        while not node_queue.empty():
            node = node_queue.get()
            node.parameters.validate()

            computation_stack.append(node)

            if node.upstream:
                for child_node in node.upstream:
                    node_queue.put(child_node)

        return computation_stack

    def _compute(self):
        computation_stack = self._build_computation_stack()

        n_step = len(self._P)

        while len(computation_stack) > 0:
            current_node = computation_stack.pop()

            if isinstance(current_node, Subbasin):
                self._logs += current_node.tank_discharge(
                    self._P, self._E, n_step, self._interval
                )

            elif isinstance(current_node, Sink) or isinstance(current_node, Junction):
                current_node.upstream_discharge(n_step)

            elif isinstance(current_node, Reach):
                self._logs += current_node.muskingum(n_step, self._interval)

            current_node.calculate_stats(self._q_obs)

    def _run(self):
        self._logs = []
        df = self._validate_dataset_n_provide_dataframe()
        df = filter_timeseries(df, self._start, self._end)
        self._init_input(df)
        self._compute()

    def reload(self):
        self._run()

    def _reconfig_by_stacked_parameters(
            self, basin_defs: List[Union[Subbasin, Reach]], stacked_parameters: List[float]
    ):
        subbasin_steps = len(SubbasinParameters().to_initial_params())
        reach_steps = len(ReachParameters().to_initial_params())
        _from = 0
        for basin_def in basin_defs:
            if isinstance(basin_def, Subbasin):
                steps = _from + subbasin_steps
                basin_def.parameters = SubbasinParameters(
                    *stacked_parameters[_from:steps]
                )
                _from = steps
            if isinstance(basin_def, Reach):
                steps = _from + reach_steps
                basin_def.parameters = ReachParameters(*stacked_parameters[_from:steps])
                _from = steps

        self._run()

    def _optimize_operator(
            self, stacked_parameters: List[float], basin_defs: List[Union[Subbasin, Reach]]
    ):
        self._reconfig_by_stacked_parameters(basin_defs, stacked_parameters)
        _nse = 0
        for basin_def in self._root_node:
            _nse += basin_def.stats.nse
        _nse = _nse / len(self._root_node)
        return 1 - _nse

    def optimize(self, basin_defs: List[Union[Subbasin, Reach]], eps: float = 0.01):
        stacked_parameters = []
        lower_bound_stacked = []
        upper_bound_stacked = []

        for basin_def in basin_defs:
            stacked_parameters.extend(basin_def.parameters.to_initial_params())
            for min_bound, max_bound in basin_def.parameters.get_bounds():
                lower_bound_stacked.append(min_bound)
                upper_bound_stacked.append(max_bound)

        initial_guess = np.array(stacked_parameters)
        param_bounds = np.column_stack((lower_bound_stacked, upper_bound_stacked))

        optimizer = minimize(
            fun=self._optimize_operator,
            x0=initial_guess,
            args=basin_defs,
            method="L-BFGS-B",
            bounds=param_bounds,
            options={"eps": eps},
        )

        self._reconfig_by_stacked_parameters(basin_defs, optimizer.x)

    def optimize_all(self, eps: float = 0.01):
        self.optimize(
            [
                basin_def
                for basin_def in self.basin_defs
                if isinstance(basin_def, (Reach, Subbasin))
            ],
            eps,
        )

    def reset_dataset(self, dataset: Dataset, interval: Optional[float]):
        self._dataset = dataset
        self._interval = interval if interval is not None else self._interval
        self._run()

    def set_timeseries_range(self, start: Optional[datetime], end: Optional[datetime]):
        self._start = start
        self._end = end
        self._run()

    def get_basin_def_by_name(self, name: str) -> Optional[BasinDef]:
        for basin_def in self._basin_defs:
            if basin_def.name == name:
                return basin_def

        raise ValueError(f"The BasinDef {name} does not exist.")

    def reconfig_parameters(self, name: str, parameters: Parameters):
        basin_def = self.get_basin_def_by_name(name)

        if (
                isinstance(basin_def, Subbasin)
                and isinstance(parameters, SubbasinParameters)
        ) or (isinstance(basin_def, Reach) and isinstance(parameters, ReachParameters)):
            basin_def.parameters = parameters

        else:
            raise ValueError(
                f"{type(parameters).__name__} cannot be configured for {type(basin_def).__name__}."
            )

        self._run()

    def _generate_plot(
            self, plot_data, xlabel="Timeseries", ylabel="Flow", figsize=(12, 6), title=None
    ):
        plt.figure(figsize=figsize)

        for label, data in plot_data:
            plt.plot(self._timeseries, data, label=label)

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

        if title:
            plt.title(title)

        plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left")
        plt.grid(True)
        plt.tight_layout()

        buf = io.BytesIO()
        plt.savefig(buf, format="PNG", bbox_inches="tight")
        buf.seek(0)
        image = Image.open(buf)
        plt.close()

        return image

    def plot_subbasin(self, subbasin: Subbasin, figsize=(12, 6)):
        if subbasin not in self._basin_defs:
            raise ValueError(f"The subbasin does not exist in the basin.")

        flow_data = [
            ("TANK 0", subbasin.q_tank_0),
            ("TANK 1", subbasin.q_tank_1),
            ("TANK 2", subbasin.q_tank_2),
            ("TANK 3", subbasin.q_tank_3),
            ("Bottom outlet flow TANK 0", subbasin.bottom_outlet_flow_tank_0),
            ("Bottom outlet flow TANK 1", subbasin.bottom_outlet_flow_tank_1),
            ("Bottom outlet flow TANK 2", subbasin.bottom_outlet_flow_tank_2),
            ("Side outlet flow TANK 1", subbasin.side_outlet_flow_tank_1),
            ("Side outlet flow TANK 2", subbasin.side_outlet_flow_tank_2),
            ("Side outlet flow TANK 3", subbasin.side_outlet_flow_tank_3),
        ]

        return self._generate_plot(
            flow_data,
            figsize=figsize,
            title=f"Subbasin {subbasin.name} Flow Characteristics",
        )

    def _plot_basin_q(self, include_all=True):
        plot_data = [("Q_obs", self._q_obs)]

        basin_defs = (
            self._root_node
            if include_all
            else [b for b in self._basin_defs if not b.upstream]
        )

        plot_data.extend(
            [(basin_def.name, basin_def.q_sim) for basin_def in basin_defs]
        )

        title = (
            "Basin Discharge (All Basins)"
            if include_all
            else "Headwater Basin Discharge"
        )
        return self._generate_plot(plot_data, ylabel="Q", title=title)

    def plot_q(self):
        return self._plot_basin_q(include_all=True)

    def plot_headwater_q(self):
        return self._plot_basin_q(include_all=False)

    def plot_basin_network(
            self,
            layout_type: Literal[
                "hierarchical", "circular", "spring", "kamada-kawai", "multipartite"
            ] = "hierarchical",
            figsize=(15, 10),
            node_spacing=1.0,
    ):
        color_map = {
            "Subbasin": "#7FB3D5",
            "Junction": "#F4D03F",
            "Reach": "#2ECC71",
            "Sink": "#E74C3C",
        }
        shape_map = {"Subbasin": "s", "Junction": "o", "Reach": "^", "Sink": "v"}

        graph = nx.DiGraph()

        def build_network(_basin_def: BasinDef):
            for us_basin_def in _basin_def.upstream:
                build_network(us_basin_def)
                graph.add_edge(us_basin_def.name, _basin_def.name)
            graph.add_node(_basin_def.name, component_type=type(_basin_def).__name__)

        for basin_def in self._root_node:
            build_network(basin_def)

        plt.figure(figsize=figsize)

        pos = None
        if layout_type == "hierarchical":
            pos = nx.spring_layout(graph, k=node_spacing)
            root_nodes = [n for n in graph.nodes() if graph.in_degree(n) == 0]
            for node in nx.descendants(graph, root_nodes[0]):
                level = nx.shortest_path_length(graph, root_nodes[0], node)
                pos[node] = (pos[node][0], 1 - level * 0.2)

        elif layout_type == "circular":
            pos = nx.circular_layout(graph)

        elif layout_type == "spring":
            pos = nx.spring_layout(graph, k=node_spacing, iterations=50)

        elif layout_type == "kamada-kawai":
            pos = nx.kamada_kawai_layout(graph)

        elif layout_type == "multipartite":
            layers = {}
            for node in graph.nodes():
                if graph.in_degree(node) == 0:
                    layers[node] = 0
                elif graph.out_degree(node) == 0:
                    layers[node] = 2
                else:
                    layers[node] = 1
            nx.set_node_attributes(graph, layers, "layer")
            pos = nx.multipartite_layout(graph, subset_key="layer")

        for component_type in ["Subbasin", "Junction", "Reach", "Sink"]:
            nodes = [
                node
                for node, attr in graph.nodes(data=True)
                if attr.get("component_type") == component_type
            ]

            if nodes:
                nx.draw_networkx_nodes(
                    graph,
                    pos,
                    nodelist=nodes,
                    node_color=color_map[component_type],
                    node_shape=shape_map[component_type],
                    node_size=1000,
                    label=component_type,
                )

        nx.draw_networkx_edges(
            graph,
            pos,
            edge_color="gray",
            arrows=True,
            arrowsize=20,
            edge_cmap=plt.cm.Blues,
            width=2,
            connectionstyle="arc3,rad=0.2",
        )

        labels = nx.draw_networkx_labels(graph, pos)
        for label in labels.values():
            label.set_bbox(dict(facecolor="white", edgecolor="none", alpha=0.7))

        plt.title(
            f"Watershed Network Visualization\n({layout_type.title()} Layout)",
            size=16,
            pad=20,
        )
        plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))
        plt.axis("off")
        plt.tight_layout()

        buf = io.BytesIO()
        plt.savefig(buf, format="PNG", bbox_inches="tight")
        buf.seek(0)

        image = Image.open(buf)
        plt.close()

        return image

    def to_dataframe(self):
        data = {"Timeseries": self._timeseries, "Q_obs": self._q_obs}
        data.update({basin_def.name: basin_def.q_sim for basin_def in self._basin_defs})
        return pd.DataFrame(data)



from typing import Optional, List

from hydtank.basin_def import BasinDef
from hydtank.bsd_flow_node import FlowNode
from hydtank.parameters import JunctionParameters


class Junction(FlowNode):
    def __init__(
            self, name: str,
            downstream: Optional[BasinDef] = None,
            upstream: Optional[List[BasinDef]] = None,
            parameters: JunctionParameters = JunctionParameters()
    ):
        super().__init__(name, downstream, upstream)
        self._parameters = parameters



import copy
import dataclasses
import json
from abc import ABC

import numpy as np


class Parameters(ABC):
    @staticmethod
    def get_bounds():
        pass

    def to_initial_params(self):
        pass

    def from_params(self, params):
        field_names = [field.name for field in dataclasses.fields(self)]
        bounds = self.get_bounds()

        for i, (name, bound) in enumerate(zip(field_names, bounds)):
            value = params[i]
            if not (bound[0] <= value <= bound[1]):
                raise ValueError(f"Value {value} for '{name}' is out of bounds {bound}")
            setattr(self, name, value)

        self.validate()

    def validate(self):
        params = self.to_initial_params()
        field_names = [field.name for field in dataclasses.fields(self)]

        for i, (param, bound) in enumerate(zip(params, self.get_bounds())):
            if not (bound[0] <= param <= bound[1]):
                raise ValueError(f"Attribute '{field_names[i]}' with value {param} is out of bounds {bound}")

    def copy(self):
        return copy.deepcopy(self)

    def __post_init__(self):
        self.validate()

    def __str__(self):
        return json.dumps(self.__dict__, indent=4)


@dataclasses.dataclass
class SubbasinParameters(Parameters):
    t0_is: float = 0.01
    t0_boc: float = 0.1
    t0_soc_uo: float = 0.1
    t0_soc_lo: float = 0.1
    t0_soh_uo: float = 75.0
    t0_soh_lo: float = 0.0
    t1_is: float = 0.01
    t1_boc: float = 0.01
    t1_soc: float = 0.01
    t1_soh: float = 0.0
    t2_is: float = 0.01
    t2_boc: float = 0.01
    t2_soc: float = 0.01
    t2_soh: float = 0.0
    t3_is: float = 0.01
    t3_soc: float = 0.01

    @staticmethod
    def get_bounds():
        bounds = (
            (0.01, 100),  # t0_is
            (0.01, 0.5),  # t0_boc
            (0.01, 0.5),  # t0_soc_uo
            (0.01, 0.5),  # t0_soc_lo
            (0, 100),  # t0_soh_uo
            (0, 50),  # t0_soh_lo
            (0.01, 100),  # t1_is
            (0.01, 0.5),  # t1_boc
            (0.01, 0.5),  # t1_soc
            (0, 100),  # t1_soh
            (0.01, 100),  # t2_is
            (0.01, 0.5),  # t2_boc
            (0.01, 0.5),  # t2_soc
            (0, 100),  # t2_soh
            (0.01, 100),  # t3_is
            (0.01, 0.5),  # t3_soc
        )
        return bounds

    def to_initial_params(self):
        return np.array([
            self.t0_is,
            self.t0_boc,
            self.t0_soc_uo,
            self.t0_soc_lo,
            self.t0_soh_uo,
            self.t0_soh_lo,
            self.t1_is,
            self.t1_boc,
            self.t1_soc,
            self.t1_soh,
            self.t2_is,
            self.t2_boc,
            self.t2_soc,
            self.t2_soh,
            self.t3_is,
            self.t3_soc,
        ])


@dataclasses.dataclass
class JunctionParameters(Parameters):

    @staticmethod
    def get_bounds():
        return ()

    def to_initial_params(self):
        return np.array([])


@dataclasses.dataclass
class SinkParameters(Parameters):

    @staticmethod
    def get_bounds():
        return ()

    def to_initial_params(self):
        return np.array([])


@dataclasses.dataclass
class ReachParameters(Parameters):
    k: float = 0.01
    x: float = 0.01

    @staticmethod
    def get_bounds():
        bounds = (
            (0.01, 5.0),  # k
            (0.01, 0.5)  # x
        )
        return bounds

    def to_initial_params(self):
        return np.array([
            self.k,
            self.x,
        ])



import numpy as np

from hydtank.basin_def import BasinDef


class FlowNode(BasinDef):
    def upstream_discharge(self, n_step: int):
        sum_node = np.zeros(n_step, dtype=np.float64)

        for us_node in self._upstream:
            sum_node += us_node.q_sim

        self._q_sim = sum_node



"""
Basin Processing Module

This module provides functionality for processing basin definition files and constructing
basin hierarchies. It handles parsing of basin configuration files and creates appropriate
basin definition objects with their relationships.
"""

from typing import List, Optional, Dict, Tuple, Any
from hydtank.basin_def import BasinDef
from hydtank.bsd_junction import Junction
from hydtank.bsd_reach import Reach
from hydtank.bsd_sink import Sink
from hydtank.bsd_subbasin import Subbasin
from hydtank.parameters import ReachParameters


def extract_basin_file_data(content: str) -> List[Dict[str, str]]:
    """
    Parse raw basin file content into a list of dictionaries containing section data.

    Args:
        content (str): Raw content of the basin definition file

    Returns:
        List[Dict[str, str]]: List of dictionaries where each dictionary represents
        a section with key-value pairs

    Example:
        >>> content = '''
        ... Subbasin: Sub1
        ... Area: 100
        ... End:
        ... '''
        >>> extract_basin_file_data(content)
        [{'Subbasin': 'Sub1', 'Area': '100'}]
    """
    sections = content.split("End:\n")
    data = []

    for section in sections:
        if not section.strip():
            continue

        lines = section.strip().split("\n")
        section_data: Dict[str, str] = {}
        current_key = None

        for line in lines:
            if ":" in line:
                key, value = map(str.strip, line.split(":", 1))
                section_data[key] = value
                current_key = key
            elif line.strip() and current_key:
                section_data[current_key] += f" {line.strip()}"

        if section_data:
            data.append(section_data)

    return data


def extract_basin_defs_dict(basin_data: List[Dict[str, str]]) -> Dict[str, Tuple[BasinDef, Optional[str]]]:
    """
    Convert parsed basin data into basin definition objects with their downstream connections.

    Args:
        basin_data (List[Dict[str, str]]): List of dictionaries containing basin section data

    Returns:
        Dict[str, Tuple[BasinDef, Optional[str]]]: Dictionary mapping basin names to tuples
        containing the basin definition object and its downstream connection name

    Raises:
        ValueError: If required parameters are missing or invalid
    """
    basin_defs_dict: Dict[str, Tuple[BasinDef, Optional[str]]] = {}
    basin_types = {'Subbasin', 'Reach', 'Junction', 'Sink'}

    for section in basin_data:
        # Find basin type and name
        basin_def_type = next((k for k in section if k in basin_types), None)
        if not basin_def_type:
            continue

        basin_def_name = section[basin_def_type]
        params = {
            'area': float(section.get('Area', 0.0)),
            'downstream': section.get('Downstream'),
            'mk': float(section.get('Muskingum K', 0.0)),
            'mx': float(section.get('Muskingum x', 0.0))
        }

        # Create appropriate basin definition object
        basin_def = _create_basin_def(basin_def_type, basin_def_name, params)
        if basin_def:
            basin_defs_dict[basin_def.name] = (basin_def, params['downstream'])

    return basin_defs_dict


def _create_basin_def(def_type: str, name: str, params: Dict[str, Any]) -> Optional[BasinDef]:
    """
    Create a specific basin definition object based on type and parameters.

    Args:
        def_type (str): Type of basin definition
        name (str): Name of the basin
        params (Dict[str, Any]): Parameters for basin creation

    Returns:
        Optional[BasinDef]: Created basin definition object or None if type is invalid
    """
    if def_type == 'Subbasin':
        return Subbasin(name, params['area'])
    elif def_type == 'Reach':
        return Reach(name, parameters=ReachParameters(params['mk'], params['mx']))
    elif def_type == 'Junction':
        return Junction(name)
    elif def_type == 'Sink':
        return Sink(name)
    return None


def build_basin_defs(basin_defs_dict: Dict[str, Tuple[BasinDef, Optional[str]]]) -> List[BasinDef]:
    """
    Build connections between basin definitions based on downstream relationships.

    Args:
        basin_defs_dict (Dict[str, Tuple[BasinDef, Optional[str]]]): Dictionary of basin
        definitions and their downstream connections

    Returns:
        List[BasinDef]: List of all basin definition objects with established connections

    Raises:
        KeyError: If a downstream reference points to a non-existent basin
    """
    for basin_def, downstream in basin_defs_dict.values():
        if downstream is not None:
            try:
                basin_def.downstream = basin_defs_dict[downstream][0]
            except KeyError:
                raise KeyError(f"Referenced downstream basin '{downstream}' does not exist")

    return [basin_def for basin_def, _ in basin_defs_dict.values()]


def build_root_node(basin_defs: List[BasinDef]) -> List[BasinDef]:
    """
    Identify root nodes and establish upstream connections in the basin hierarchy.

    Args:
        basin_defs (List[BasinDef]): List of basin definitions to process

    Returns:
        List[BasinDef]: List of root nodes (basins with no downstream connections)

    Example:
        >>> sub1 = Subbasin("Sub1", 100)
        >>> sub2 = Subbasin("Sub2", 200)
        >>> junction = Junction("J1")
        >>> sub1.downstream = junction
        >>> sub2.downstream = junction
        >>> root_nodes = build_root_node([sub1, sub2, junction])
        >>> len(root_nodes)
        1
        >>> root_nodes[0].name
        'J1'
    """
    root_nodes: List[BasinDef] = []

    for basin_def in basin_defs:
        if basin_def.downstream is None:
            root_nodes.append(basin_def)
        else:
            downstream = basin_def.downstream
            if not downstream.upstream:
                downstream.upstream = [basin_def]
            else:
                downstream.upstream.append(basin_def)

    return root_nodes



import json
from typing import Optional, List

import numpy as np

from hydtank.parameters import Parameters
from hydtank.statistics import Statistics


class BasinDef:
    def __init__(
            self, name: str,
            downstream: Optional['BasinDef'] = None,
            upstream: Optional[List['BasinDef']] = None,
    ):
        self._name = name
        self._parameters = Parameters()
        self._stats = Statistics()
        self._downstream = downstream
        self._upstream = [] if upstream is None else upstream
        self._q_sim: Optional[np.ndarray] = None

    def calculate_stats(self, q_obs: np.ndarray):
        self._stats.calculate_stats(q_obs, self._q_sim)

    def display(self, visited=None, level=0):
        if visited is None:
            visited = set()
        if self in visited:
            print('  ' * level + f"(Visited) {self._name}")
            return
        visited.add(self)

        print('  ' * level + f"{type(self).__name__}: {self._name}")

        if self._upstream:
            print('  ' * level + "Upstream:")
            for upstream in self._upstream:
                upstream.display(visited, level + 1)

        if self._downstream:
            print('  ' * level + "Downstream:")
            self._downstream.display(visited, level + 1)

    @property
    def name(self) -> str:
        return self._name

    @property
    def downstream(self) -> Optional['BasinDef']:
        return self._downstream

    @downstream.setter
    def downstream(self, value: Optional['BasinDef']):
        self._downstream = value

    @property
    def upstream(self) -> List['BasinDef']:
        return self._upstream

    @upstream.setter
    def upstream(self, value: List['BasinDef']):
        self._upstream = value

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, value: Parameters):
        self._parameters = value

    @property
    def q_sim(self):
        return self._q_sim

    @property
    def stats(self):
        return self._stats

    def __str__(self):
        return json.dumps(
            {
                "Name": self.name,
                "Downstream": None if self._downstream is None else self._downstream.name,
                "Upstreams": [basin_def.name for basin_def in self._upstream],
                "Parameters": self._parameters.__dict__
            },
            indent=4
        )



import copy
from datetime import datetime
from typing import List

from hydutils.hyd_constants import (
    TIMESERIES,
    PRECIPITATION,
    EVAPOTRANSPIRATION,
    DISCHARGE,
)
import pandas as pd


class Dataset:
    def __init__(
        self,
        time_series: List[datetime],
        precipitation: List[float],
        evapotranspiration: List[float],
        discharge: List[float],
    ):
        self._time_series = time_series
        self._precipitation = precipitation
        self._evapotranspiration = evapotranspiration
        self._discharge = discharge

    def to_dataframe(self):
        dataset_dict = {
            TIMESERIES: self._time_series,
            PRECIPITATION: self._precipitation,
            EVAPOTRANSPIRATION: self._evapotranspiration,
            DISCHARGE: self._discharge,
        }

        df = pd.DataFrame(dataset_dict)
        df[TIMESERIES] = pd.to_datetime(df[TIMESERIES])
        return df

    def copy(self):
        return copy.deepcopy(self)
